#include "Heap.h"

/* peneira a heap fazendo o elemento i descer para a posio em que no possua filho maior */
void peneiraMaxHeap(int n, NoHeap lista[], int i){
    
    /* filhos */
    int fesq = 2*(i+1)-1;
    int fdir = 2*(i+1);
    
    /* identifica o maior elemento entre i e os seu filhos, caso existam */
    int maior = i;
    
    /* como o mtodo  recursivo enquanto no identicar uma posio estvel */
    /* a cada passo o piv i  trocado por um elemento mais distante da raiz */
    /* precisa verificar se o novo elemento no  um n folha */
    /* o elemento apenas possuir filho se a posio do filho pertence a heap */
    /* a cada chamada externa da heap ela foi reduzida de tamanho */
    if (fesq < n && lista[fesq].val > lista[maior].val){   
        maior = fesq;
    }
    
    if (fdir < n && lista[fdir].val > lista[maior].val){
        maior = fdir;
    }
    
    /* se o maior elemento no for o piv i ele est fora da posio desejada */
    if (maior != i){
    
        NoHeap aux;        
        
        /* troca o piv de posio com o maior elemento */
        aux = lista[maior];
        lista[maior] = lista[i];
        lista[i] = aux;

        /* chamada recursiva com o novo piv (antiga posio do maior filho) */
        peneiraMaxHeap(n, lista, maior);
    }
}

/* peneira a heap fazendo o elemento i descer para a posio em que no possua filho menor */
void peneiraMinHeap(int n, NoHeap lista[], int i){
    
    /* filhos */
    int fesq = 2*(i+1)-1;
    int fdir = 2*(i+1);
    
    /* identifica o menor elemento entre i e os seu filhos, caso existam */
    int menor = i;
    
    /* como o mtodo  recursivo enquanto no identicar uma posio estvel */
    /* a cada passo o piv i  trocado por um elemento mais distante da raiz */
    /* precisa se verificar se o novo elemento no  um n folha */
    /* o elemento apenas possuir filho se a posio do filho pertence a heap */
    /* a cada chamada externa da heap ela foi reduzida de tamanho */
    if (fesq < n && lista[fesq].val < lista[menor].val){   
        menor = fesq;
    }
    
    if (fdir < n && lista[fdir].val < lista[menor].val){
        menor = fdir;
    }
    
    /* se o menor elemento no for o piv i ele est fora da posio desejada */
    if (menor != i){
    
        NoHeap aux;        
        
        /* troca o piv de posio com o menor elemento */
        aux = lista[menor];
        lista[menor] = lista[i];
        lista[i] = aux;

        /* chamada recursiva com o novo piv (antiga posio do menor filho) */
        peneiraMinHeap(n, lista, menor);
    }
}

/* constroi uma heap */
/* numa min-heap cada n pai no pode possuir filho menor; a raiz  o menor elemento da heap */
void constroiMinHeap(int n, NoHeap lista[]){
    
    int i;
    
    /* a partir do pai com maior indice (o mais distante da raiz) */
    for (i = (int)(n-2)/2; i >= 0; i--){
        /* peneira a heap fazendo o elemento da raiz descer para a posio de estabilidade */
        peneiraMinHeap(n, lista, i);  
    }
}

/* constroi uma heap */
/* numa max-heap cada n pai no pode possuir filho maior; a raiz  o maior elemento da heap */
void constroiMinHeap(int n, NoHeap lista[]){
    
    int i;
    
    /* a partir do pai com maior indice (o mais distante da raiz) */
    for (i = (int)(n-2)/2; i >= 0; i--){
        /* peneira a heap fazendo o elemento da raiz descer para a posio de estabilidade */
        peneiraMaxHeap(n, lista, i);
    }
}

/* monta uma estrutura de heap */
/* a cada iterao troca o elemento da raiz da heap com o ultimo elemento da lista e exclui o ultimo elemento da heap */
/* ento rearruma a heap */
void minHeapSort(int n, NoHeap lista[]){
 
    /* m  o tamanho da heap */
    /* o tamanho da lista ordenada  n-m */
    int m
	NoHeap aux;

    /* constroi uma heap */    
    constroiMinHeap(n, lista);
        
    /* a cada decremento diminui o tamanho da heap excluindo o ultimo elemento */
    /* adiciona o elemento a posio mais recente da lista */
    for (m = n-1; m > 0; m--){
     
        /* troca a raiz da heap com o ultimo elemento da heap */
        aux = lista[m];
        lista[m] = lista [0];
        lista[0] = aux;

        /* peneira a heap fazendo o elemento da raiz descer para a posio de estabilidade */
        peneiraMinHeap(m, lista, 0);
		    
    }
}


/* monta uma estrutura de heap */
/* a cada iterao troca o elemento da raiz da heap com o ultimo elemento da lista e exclui o ultimo elemento da heap */
/* ento rearruma a heap */
void maxHeapSort(int n, NoHeap lista[]){
 
    /* m  o tamanho da heap */
    /* o tamanho da lista ordenada  n-m */
    int m;
	NoHeap aux;

    /* constroi uma heap */    
    constroiMaxHeap(n, lista);
        
    /* a cada decremento diminui o tamanho da heap excluindo o ultimo elemento */
    /* adiciona o elemento a posio mais recente da lista */
    for (m = n-1; m > 0; m--){
     
        /* troca a raiz da heap com o ultimo elemento da heap */
        aux = lista[m];
        lista[m] = lista [0];
        lista[0] = aux;

        /* peneira a heap fazendo o elemento da raiz descer para a posio de estabilidade */

        peneiraMaxHeap(m, lista, 0);        
    }
}
